//region Helper functions
function safe_key_union(sourceA, sourceB) {
    const aBad = sourceA === undefined || sourceA === null;
    const bBad = sourceB === undefined || sourceB === null;

    if (aBad && bBad) return new Set();

    if (aBad) {
        const bKeys = Object.keys(sourceB);
        return new Set(bKeys);
    }
    const aKeys = Object.keys(sourceA);

    if (bBad) {
        return new Set(aKeys);
    }
    const bKeys = Object.keys(sourceB);

    let setA = new Set(aKeys);
    let setB = new Set(bKeys);

    for (let elem of setB) {
        setA.add(elem);
    }
    return setA;
}

//endregion
/*
var triangulationParams = {
    accuracy: 1,    		// float between 0 and 1
    blur: 0,         		// positive integer
    threshold: 50,    		// integer between 1 and 100
    vertexCount: 14000, 		// positive integer
    fill: '#000000',       		// boolean or string with css color (e.g '#bada55', 'red', rgba(100,100,100,0.5))
    stroke: true,     		// boolean or string with css color (e.g '#bada55', 'red', hsla(0, 50%, 52%, 0.5))
    strokeWidth: 0.5, 		// positive float
    gradients: false,  		// boolean
    gradientStops: 4, 		// positive integer >= 2
    lineJoin: 'miter',		// 'miter', 'round', or 'bevel'
    transparentColor: false // boolean false or string with css color
};
*/
class Mask {

    constructor() {
        this.container = {};
    }

    setRed(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] |= 0x00FF0000;
    }

    setGreen(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] |= 0x0000FF00;
    }

    setBlue(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] |= 0x000000FF;
    }

    clearRed(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] &= ~0x00FF0000;
    }

    clearGreen(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] &= ~0x0000FF00;
    }

    clearBlue(x, y) {
        this.ensureAvailable(x, y);
        this.container[x][y] &= ~0x000000FF;
    }

    testRed(x, y) {
        return x in this.container && y in this.container[x] && (this.container[x][y] & 0x00FF0000) !== 0;
    }

    testGreen(x, y) {
        return x in this.container && y in this.container[x] && (this.container[x][y] & 0x0000FF00) !== 0;
    }

    testBlue(x, y) {
        return x in this.container && y in this.container[x] && (this.container[x][y] & 0x000000FF) !== 0;
    }

    ensureAvailable(x, y) {
        if (!(x in this.container)) {
            this.container[x] = {};
        }
        if (!(y in this.container[x])) {
            this.container[x][y] = 0xFF000000;
        }
    }

    getMask(x, y) {
        if (!(x in this.container)) return null;
        if (!(y in this.container[x])) return null;
        return this.container[x][y];

    }

    setMask(xKey, yKey, value) {
        this.ensureAvailable(xKey, yKey);
        this.container[xKey][yKey] = value;
    }

    union(that) {
        let result = new Mask()

        let xKeys = safe_key_union(this.container, that.container);
        for (const xKey of xKeys) {
            let yKeys = safe_key_union(this.container[xKey], that.container[xKey]);
            for (const yKey of yKeys) {
                const thisValue = this.getMask(xKey, yKey);
                const thatValue = that.getMask(xKey, yKey);

                let value;
                if (thisValue === null || thisValue === undefined) {
                    value = thatValue;
                } else if (thatValue === null || thatValue === undefined) {
                    value = thisValue;
                } else {
                    value = thisValue | thatValue;
                }
                result.setMask(xKey, yKey, value);
            }
        }

        return result;
    }

    calcWidth() {
        let min_x = Number.MAX_VALUE;
        let max_x = -Number.MAX_VALUE;
        for (let x in this.container) {
            if (!this.container.hasOwnProperty(x)) continue;
            x = parseInt(x);
            if (x < min_x) {
                min_x = x;
            }
            if (x > max_x) {
                max_x = x;
            }
        }
        let d_x = max_x - min_x;
        return d_x;
    }

    calcHeight() {
        let min_y = Number.MAX_VALUE;
        let max_y = -Number.MAX_VALUE;
        for (let x in this.container) {
            if (!this.container.hasOwnProperty(x)) continue;
            x = parseInt(x);

            for (let y in this.container[x]) {
                if (!this.container[x].hasOwnProperty(y)) continue;
                y = parseInt(y);

                if (y < min_y) {
                    min_y = y;
                }
                if (y > max_y) {
                    max_y = y;
                }
            }
        }
        let d_y = max_y - min_y;
        return d_y;
    }
}

class Rectangle {
    constructor(x, y, w, h) {
        this.x = x;
        this.y = y;
        this.w = w;
        this.h = h;
    }

    coordinateInside(x, y) {
        return !(x < this.x || y < this.y || x >= this.x + this.w || y >= this.y + this.h);
    }

    area() {
        return this.w * this.h;
    }

    scale(w_scalar, h_scalar) {
        const width_increase = Math.ceil(this.w * w_scalar) - this.w;
        const height_increase = Math.ceil(this.h * h_scalar) - this.h;

        this.x -= Math.ceil(width_increase / 2.0);
        this.y -= Math.ceil(height_increase / 2.0);

        this.w += width_increase;
        this.h += height_increase;
    }

    scaled(w_scalar, h_scalar) {
        const new_rect = this.clone();
        new_rect.scale(w_scalar, h_scalar);
        return new_rect;
    }

    stretch(width, height) {
        this.x -= width;
        this.w += 2 * width;

        this.y -= height;
        this.h += 2 * height;
    }

    stretched(width, height) {
        const new_rect = new Rectangle(this.x, this.y, this.w, this.h);
        new_rect.stretch(width, height);
        return new_rect;
    }

    resize(width, height) {
        this.x -= Math.ceil(width/2);
        this.w += width;

        this.y -= Math.ceil(height/2);
        this.h += height;
    }

    resized(width, height) {
        const new_rect = new Rectangle(this.x, this.y, this.w, this.h);
        new_rect.resize(width, height);
        return new_rect;
    }


    clone() {
        return new Rectangle(this.x, this.y, this.w, this.h);
    }

    global_cropped(x, y, w, h) {
        const result = this.clone();
        result.x = Math.max(result.x, x);
        result.y = Math.max(result.y, y);

        const x2 = x + w;
        const y2 = y + h;

        const result_x2 = result.x + result.w;
        const result_y2 = result.y + result.h;

        const target_result_x2 = Math.min(x2, result_x2);
        const target_result_y2 = Math.min(y2, result_y2);

        result.w = target_result_x2 - result.x;
        result.h = target_result_y2 - result.y;

        return result;

    }
}

class GlitchConfig {
    constructor() {
        this.r_channel_intensity = 0;
        this.g_channel_intensity = 0;
        this.b_channel_intensity = 0;
        this.r_negate_intensity = 0;
        this.g_negate_intensity = 0;
        this.b_negate_intensity = 0;
        this.shift_horizontal = false;
        this.shift_vertical = false;
        this.h_bar_size_range = [0, 0];
        this.v_bar_size_range = [0, 0];
        this.shift_intensity_range = [0, 0];
    }

    set_channel_intensities(value) {
        this.r_channel_intensity = value;
        this.g_channel_intensity = value;
        this.b_channel_intensity = value;
    }

    set_negation_intensities(value) {
        this.r_negate_intensity = value;
        this.g_negate_intensity = value;
        this.b_negate_intensity = value;
    }

    set_all_intensities(value) {
        this.set_channel_intensities(value);
        this.set_negation_intensities(value);
    }
}

class ImageProcessor {
    constructor(original, configurations) {

        //this.sum = 0;
        this.effects_applied = 0;
        this.marvin_out = original
        this.source = document.createElement('canvas');
        this.source.width = original.canvas.width;
        this.source.height = original.canvas.height;
        this.sourcectx = this.source.getContext('2d');
        this.sourcectx.drawImage(original.canvas, 0, 0, original.canvas.width, original.canvas.height);

        this.mask = document.createElement('canvas');
        this.mask.width = original.canvas.width;
        this.mask.height = original.canvas.height;
        this.maskctx = this.mask.getContext('2d');
        this.maskctx.fillStyle = 'rgba(0,0,0,0)';
        this.maskctx.fillRect(0, 0, original.canvas.width, original.canvas.height);
        this.maskctx.fillStyle = '#fff';


        this.bar_configuration = configurations.bar_configuration;
        this.pixel_configuration = configurations.pixel_configuration;
        this.blur_configuration = configurations.blur_configuration;
        this.glitch_configuration = configurations.glitch_configuration;
        this.triangle_configuration = configurations.triangle_configuration;
        this.sticker_configuration = configurations.sticker_configuration;
        this.sticker_collections = configurations.sticker_collections;
        this.reverse_mode_configuration = configurations.reverse_mode_configuration;
        this.natural_width = original.canvas.width;
        this.natural_height = original.canvas.height;

        this.blurred = null;
        //this.marvin_original = null;

        this.triangulized = false;
        this.sobelized = false;

        this.per_layer_queues = []
    }

    prepareMask(rect, config, smoothing = 0) {
        if (smoothing > 0) {
            this.maskctx.filter = 'blur(' + smoothing + 'px)';
        }

        switch (config._shape) {
        //switch (3) {
            case 1:
                this.maskctx.beginPath();
                this.maskctx.arc(rect.x + rect.w / 2, rect.y + rect.h / 2, Math.min(rect.w / 2, rect.h / 2), 0, 2 * Math.PI);
                this.maskctx.fill();
                break;
            case 2:
                this.maskctx.beginPath();
                this.maskctx.ellipse(rect.x + rect.w / 2, rect.y + rect.h / 2,
                    rect.w / 2,
                    rect.h / 2, 0, 0, 4 * Math.PI);
                this.maskctx.fill();
                break;
            case 3:  // Heart
                //Heart looks better when not streched so we adjust the witdh / height ratio to be 1

                let rwh = (rect.w + rect.h) / 2;
                rect.resize((rwh-rect.w), (rwh-rect.h));
                /*
                let x = rect.x+rect.w/2;
                let y = rect.y;
                let width = rect.w;//Math.min(rect.w, rect.h) ;
                let height = rect.h;

                this.maskctx.save();
                this.maskctx.beginPath();
                let topCurveHeight = height * 0.3;
                this.maskctx.moveTo(x, y + topCurveHeight);
                // top left curve
                this.maskctx.bezierCurveTo(
                    x, y,
                    x - width / 2, y,
                    x - width / 2, y + topCurveHeight
                );

                // bottom left curve
                this.maskctx.bezierCurveTo(
                    x - width / 2, y + (height + topCurveHeight) / 2,
                    x, y + (height + topCurveHeight) / 2,
                    x, y + height
                );

                // bottom right curve
                this.maskctx.bezierCurveTo(
                    x, y + (height + topCurveHeight) / 2,
                    x + width / 2, y + (height + topCurveHeight) / 2,
                    x + width / 2, y + topCurveHeight
                );

                // top right curve
                this.maskctx.bezierCurveTo(
                    x + width / 2, y,
                    x, y,
                    x, y + topCurveHeight
                );

                this.maskctx.closePath();
                //ctx.fillStyle = color;
                this.maskctx.fill();
                this.maskctx.restore();
                break;
                */

                let x = rect.x+rect.w/2;
                let y = rect.y;
                let width = rect.w;//Math.min(rect.w, rect.h) ;
                let height = rect.h;
                let topCurveHeight = height * 0.3;
                this.maskctx.save();
                this.maskctx.beginPath();
                this.maskctx.moveTo( x, y + topCurveHeight);
                this.maskctx.bezierCurveTo(
                    x, y, x - width,
                    y, x , y + height);

                this.maskctx.moveTo(  x, y + topCurveHeight);
                this.maskctx.bezierCurveTo(
                    x, y, x + width,
                    y, x , y + height);
                this.maskctx.closePath();
                this.maskctx.fill();
                this.maskctx.restore();

                break;
            default:
                //Rectangle
                this.maskctx.fillRect(rect.x, rect.y, rect.w, rect.h);
        }
    }

    pixelateArea(rect) {
        const pixel_configuration = this.pixel_configuration;

        rect = rect.scaled(pixel_configuration._scale, pixel_configuration._scale);
        if (pixel_configuration._feathering > 0) {
            let smoothing_strength = calc_smoothing(rect, pixel_configuration);
            this.prepareMask(rect, pixel_configuration, smoothing_strength);
            rect.resize(smoothing_strength * 2, smoothing_strength * 2);
        } else {
            this.prepareMask(rect, pixel_configuration);
        }

        if (!this.reverse_mode_configuration.enabled) {
            let distance = pixel_configuration._distance;
            let box_size = pixel_configuration._pixel_size;
            // TODO fix _geometry to be an int (or other non magic-number value)
            let geometry = parseInt(pixel_configuration._geometry);
            if (pixel_configuration._pixel_size_mode === 0) {
                const smallest_dimension = Math.min(this.natural_height, this.natural_width);
                box_size = Math.ceil(smallest_dimension / 100 * pixel_configuration._pixel_size);
            }
            // Hexelate
            if (geometry === 1) {
                hexelate_area(this.sourcectx, rect, box_size, distance, false, this.marvin_out);
                // Hexelate glitch
            } else if (geometry === 2) {
                hexelate_area(this.sourcectx, rect, box_size, distance, true, this.marvin_out);
            } else {
                pixelate_rectangle(this.sourcectx, rect, box_size, distance, this.marvin_out);
            }
        }
    }

    blurArea(rect) {
        const blur_configuration = this.blur_configuration;
        rect = rect.scaled(blur_configuration._scale, blur_configuration._scale);
        if (this.blurred == null) {
            this.marvin_out.update();
            this.blurred = generateBaseBlurImproved(this.marvin_out.canvas, blur_configuration._blur_strength);
        }
        if (blur_configuration._feathering > 0) {
            let smoothing_strength = calc_smoothing(rect, blur_configuration);
            this.prepareMask(rect, blur_configuration, smoothing_strength);
            rect.resize(smoothing_strength * 2, smoothing_strength * 2);
        } else {
            this.prepareMask(rect, blur_configuration);
        }
        if (!this.reverse_mode_configuration.enabled) {
            this.sourcectx.drawImage(this.blurred, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
        }
    }

    triangulizeArea(rect){
        const triangulize_configuration = this.triangle_configuration;
        rect = rect.scaled(triangulize_configuration._scale, triangulize_configuration._scale);

        if (triangulize_configuration._feathering > 0) {
            let smoothing_strength = calc_smoothing(rect, triangulize_configuration);
            this.prepareMask(rect, triangulize_configuration, smoothing_strength);
            rect.resize(smoothing_strength * 2, smoothing_strength * 2);
        } else {
            this.prepareMask(rect, triangulize_configuration);
        }

        if (!this.reverse_mode_configuration.enabled && !this.triangulized) {
            this.triangulized = true;
            let imageData = this.sourcectx.getImageData( 0, 0, this.source.width,  this.source.height );
            let vtc = triangulize_configuration.vertexCount;
            if(triangulize_configuration.vertexMode){
                vtc = Math.ceil(((this.source.width+this.source.height)/100)*triangulize_configuration.vertexPercentage);
            }
            let triangulationParams = {
                accuracy: triangulize_configuration.accuracy,
                blur: triangulize_configuration.blur,
                threshold: triangulize_configuration.threshold,
                vertexCount: vtc,
                fill: triangulize_configuration.fill,
                stroke: triangulize_configuration.stroke,
                strokeWidth: triangulize_configuration.strokeWidth,
                gradients: triangulize_configuration.gradients,
                gradientStops: triangulize_configuration.gradientStops,
                lineJoin: triangulize_configuration.lineJoin,
                transparentColor: triangulize_configuration.transparentColor,
            };
            let triangulizedImageData = triangulate(triangulationParams)
                .fromImageDataSync( imageData )
                .toImageDataSync();
            this.sourcectx.putImageData( triangulizedImageData, 0, 0 );
        }

    }

    blackenArea(rect) {
        const bar_configuration = this.bar_configuration;
        rect = rect.scaled(bar_configuration._scale, bar_configuration._scale);

        if (bar_configuration._feathering > 0) {
            let smoothing_strength = calc_smoothing(rect, bar_configuration);
            this.prepareMask(rect, bar_configuration, smoothing_strength);
            rect.resize(smoothing_strength * 2, smoothing_strength * 2);
        } else {
            this.prepareMask(rect, bar_configuration);
        }

        // If not inverse
        if (!this.reverse_mode_configuration.enabled) {
            this.sourcectx.fillStyle = bar_configuration._color;
            this.sourcectx.fillRect(rect.x, rect.y, rect.w, rect.h);
        }
    }

    sobelArea(rect){
        const bar_configuration = this.bar_configuration;
        rect = rect.scaled(bar_configuration._scale, bar_configuration._scale);

        if (bar_configuration._feathering > 0) {
            let smoothing_strength = calc_smoothing(rect, bar_configuration);
            this.prepareMask(rect, bar_configuration, smoothing_strength);
            rect.resize(smoothing_strength * 2, smoothing_strength * 2);
        } else {
            this.prepareMask(rect, bar_configuration);
        }

        // If not inverse
        if (!this.reverse_mode_configuration.enabled) {
            /*
            let t0 = performance.now();
            let imageData = this.sourcectx.getImageData( rect.x, rect.y, rect.w,  rect.h );
            let sobelData = Sobel(imageData);
            let sobelImageData = sobelData.toImageData();
            this.sourcectx.putImageData(sobelImageData,rect.x, rect.y);
            let t1 = performance.now();
            console.log("Effect took " + (t1 - t0) + " ms.");
            */
            if(this.sobelized == false){
                let t0 = performance.now();
                this.sobelized = true;
                let imageData = this.sourcectx.getImageData( 0,0, this.source.width,  this.source.height );
                let sobelData = Sobel(imageData);
                let sobelImageData = sobelData.toImageData();
                this.sourcectx.putImageData(sobelImageData,0,0);
                let t1 = performance.now();
                console.log("Sobel took " + (t1 - t0) + " ms.");
            }

        }
    }

    applySticker(klass, rect, cache) {
        const sticker_configuration = this.sticker_configuration;
        rect = rect.scaled(sticker_configuration._scale, sticker_configuration._scale);

        let sticker_collections = this.sticker_collections.filter(obj => {
            return obj._enabled;
        });
        if(sticker_collections.length == 0){
            return;
        }
        let stickers = sticker_collections[0]._stickers;


        if(sticker_configuration._consistency && 'sticker_consistency_'+klass.key in cache){
            stickers = cache['sticker_consistency_'+klass.key];
        }else if(sticker_configuration._draw_mode == 0){
            //Completely random, every box a sticker from any collection
            stickers = sticker_collections[Math.floor(Math.random() * sticker_collections.length)]._stickers;
        }else if(sticker_configuration._draw_mode == 1){
            // One collection per image
            let index = 0;
            if('sticker_draw_collection' in cache){
                index = cache.sticker_draw_collection;
            }else{
                index = Math.floor(Math.random() * sticker_collections.length);
                cache.sticker_draw_collection = index;
            }
            stickers = sticker_collections[index]._stickers;
        }else if(sticker_configuration._draw_mode == 2){
            stickers = [];
            sticker_collections.forEach(function(c){
                stickers = stickers.concat(getStickersByGroup(c,sticker_configuration._groups));
            });
        }
        //TODO: If -1 klass because of merge, try to find sticker which matches all first
        if(klass.index >= 0){
            stickers = stickers.filter(sticker => {
                return sticker.klasses.some(clas => clas.key === klass.key)
            });
        }
        if(stickers.length == 0){
            return;
        }
        let sticker = this.draw_sticker(stickers);
        if(sticker_configuration._consistency && !('sticker_consistency_'+klass.key in cache)){
            cache['sticker_consistency_'+klass.key] = [sticker];
        }

        rect = rect.scaled(sticker.scale, sticker.scale);
        let buffer = sticker.file;
        const canvasElement = document.createElement('canvas');
        canvasElement.width = sticker.width;
        canvasElement.height = sticker.height;
        const ctx = canvasElement.getContext("2d");
        let imageData = ctx.createImageData(canvasElement.width, canvasElement.height);
        imageData.data.set(buffer);
        ctx.putImageData(imageData,0,0);
        let hRatio = rect.w / canvasElement.width;
        let vRatio = rect.h / canvasElement.height;
        let ratio  = Math.max ( hRatio, vRatio );
        let w_adjusted = canvasElement.width*ratio;
        let h_adjusted = canvasElement.height*ratio;
        let x_adjusted = rect.x-(Math.abs(rect.w-w_adjusted)/2);
        let y_adjusted = rect.y-(Math.abs(rect.h-h_adjusted)/2);
        rect = new Rectangle(x_adjusted, y_adjusted, w_adjusted, h_adjusted)
        if (sticker_configuration._feathering > 0) {
            let smoothing_strength = calc_smoothing(rect, sticker_configuration);
            this.prepareMask(rect, sticker_configuration, smoothing_strength);
            rect.resize(smoothing_strength * 2, smoothing_strength * 2);
        } else {
            this.prepareMask(rect, sticker_configuration);
        }

        // If not inverse
        if (!this.reverse_mode_configuration.enabled) {
            this.sourcectx.drawImage(canvasElement, rect.x, rect.y, rect.w, rect.h);
        }
    }

    draw_sticker(stickers) {
         if(stickers.length == 0){
            return null;
         }

         let i;
         let weights = [];
         for (i = 0; i < stickers.length; i++){
             if(!stickers[i].chance && stickers[i].chance != 0){
                 stickers[i].chance = 1;
             }
             weights[i] = stickers[i].chance + (weights[i - 1] || 0);
         }
         let random = Math.random() * weights[weights.length - 1];
         for (i = 0; i < weights.length; i++){
             if (weights[i] > random) {
                 break;
             }
         }
         return stickers[i];
    }

    drawTextBox(ctx, x, y, text, bg_color, baseline = "top", font_sz = 16, font_family = "sans-serif", fg_color = "#000000") {
        ctx.font = `${font_sz}px ${font_family}`;
        ctx.textBaseline = baseline;

        const metrics = ctx.measureText(text);
        const string_width = Math.ceil(ctx.measureText(text).width);
        const font_height = font_sz

        let y_adjust = baseline === "bottom" ? -(2 + font_height) : 0;
        ctx.fillStyle = bg_color;
        ctx.fillRect(x, y + y_adjust, string_width + 4, font_height + 4);

        // text = text + `sz: ${string_width}`;
        ctx.fillStyle = fg_color;
        ctx.fillText(text, x, y);
    }

    namedBoxRect(klass, score, rect) {
        const bg_color = "#00FFFF";

        let font_size = 2 + Math.ceil((this.natural_width * 14) / 1000);
        font_size = Math.max(9, font_size);

        const ctx = this.marvin_out.ctx;

        ctx.strokeStyle = bg_color;
        ctx.lineWidth = 4;

        ctx.strokeRect(rect.x, rect.y, rect.w, rect.h);

        let text_to_write = klass.name;
        this.drawTextBox(ctx, rect.x, rect.y, text_to_write, bg_color, "top", font_size);

        text_to_write = score;
        this.drawTextBox(ctx, rect.x, rect.y + rect.h, text_to_write, bg_color, "bottom", font_size - 2);

        this.marvin_out.imageData = ctx.getImageData(0, 0, this.natural_width, this.natural_height);
    }

    fettyfiArea(klass, rect) {
        rect = rect.scaled(this.glitch_configuration._scale, this.glitch_configuration._scale);
        let effect_range = Math.ceil(0.3 * Math.min(rect.h, rect.w) * this.glitch_configuration.intensity_shift);
        let effect_bounds = rect.stretched(effect_range, effect_range);

        rect = rect.global_cropped(0, 0, this.natural_width, this.natural_height);
        effect_bounds = effect_bounds.global_cropped(0, 0, this.natural_width, this.natural_height);

        let mul = 1;
        if (klass.erotic) {
            mul = 5;
        }
        mul = Math.ceil(mul * this.glitch_configuration.intensity_color);


        const source_canvas = document.createElement('canvas');
        source_canvas.width = effect_bounds.w;
        source_canvas.height = effect_bounds.h;
        let source_ctx = source_canvas.getContext('2d');
        let sx = effect_bounds.x;
        let sy = effect_bounds.y;
        let sw = effect_bounds.w;
        let sh = effect_bounds.h;
        let dx = 0;
        let dy = 0;
        let dw = effect_bounds.w;
        let dh = effect_bounds.h;
        source_ctx.putImageData(this.sourcectx.getImageData(sx, sy, sw, sh), dx, dy);
        //We can prob. save some ms here as this.sourcectx.getImageData(sx, sy, sw, sh) (one line above) does already request all the pixels we need
        //so this is kinda redundant what we are doing here and just creates an offset where the array starts to contain pixels...
        //Also: We should do this once for the entire image and not once for each box, this takes 10ms or more to create... so we can save 100ms or more on many images
        const source_data_obj = source_ctx.getImageData(0, 0, source_canvas.width, source_canvas.height);

        let config;

        // Round 1:
        // Green, Horizontal, 50%, default bar & shift sizes
        config = new GlitchConfig();
        config.g_channel_intensity = 0.75 * mul;
        // config.set_channel_intensities(0.05 * mul);
        config.shift_horizontal = this.glitch_configuration.shift_horizontal;
        config.h_bar_size_range = [Math.floor(rect.h / 20), Math.ceil(rect.h / 10)];
        config.shift_intensity_range = [Math.round(effect_range / 4), effect_range];


        let t1_canvas = document.createElement('canvas');
        t1_canvas.width = effect_bounds.w;
        t1_canvas.height = effect_bounds.h;
        let t1_ctx = t1_canvas.getContext('2d');
        t1_ctx.putImageData(source_ctx.getImageData(dx, dy, dw, dh), dx, dy);


        let mask1 = this.neo_shift(source_ctx, source_data_obj, t1_ctx, t1_canvas, rect, effect_bounds, config);


        // Round 2:
        // Red, Vertical, 200%, default bar & shift sizes.
        config = new GlitchConfig();
        config.r_channel_intensity = 0.85 * (1 + mul);
        // config.set_channel_intensities(0.6 * mul);
        config.shift_vertical = this.glitch_configuration.shift_vertical;
        config.v_bar_size_range = [Math.floor(rect.w / 20), Math.ceil(rect.w / 10)];
        config.shift_intensity_range = [Math.round(effect_range / 12), Math.round(effect_range / 2)];

        let t2_canvas = document.createElement('canvas');
        t2_canvas.width = effect_bounds.w;
        t2_canvas.height = effect_bounds.h;
        let t2_ctx = t2_canvas.getContext('2d');
        t2_ctx.putImageData(t1_ctx.getImageData(dx, dy, dw, dh), dx, dy);

        /* Todo: enabling these lines will double the duration again, can we somehow improve this? */
        const t1_data_obj = t1_ctx.getImageData(0, 0, t1_canvas.width, t1_canvas.height);
        let mask2 = this.neo_shift(t1_ctx, t1_data_obj, t2_ctx, t2_canvas, rect, effect_bounds, config);
        let mask3 = mask1.union(mask2);
        // let mask3 = mask1;

        let area = effect_bounds.area();
        let logSize = Math.log(area);
        let sizeScalar = Math.round(logSize);


        this.enqueue(20 - sizeScalar, effect_bounds, t2_canvas, mask3);
    }

    enqueue(layer, rect, canvas, mask) {
        if (!(layer in this.per_layer_queues)) {
            this.per_layer_queues[layer] = [];
        }
        this.per_layer_queues[layer].push([rect, canvas, mask]);
    }

    applyEffect(klass, score, rect, type, cache) {
        //let t0 = performance.now();
        this.effects_applied += 1;
        switch (type) {
            case effects.PIXEL:
                this.pixelateArea(rect);
                break;
            case effects.BLUR:
                this.blurArea(rect);
                break;
            case effects.BLACK:
                this.blackenArea(rect);
                break;
            case effects.BOX:
                this.namedBoxRect(klass, score, rect);
                break;
            case effects.TRIANGLE:
                this.triangulizeArea(rect);
                break;
            case effects.GLITCH:
                this.fettyfiArea(klass, rect);
                break;
            case effects.STICKER:
                this.applySticker(klass, rect, cache);
                break;
            case effects.NONE:
                break;
        }
        //let t1 = performance.now();
        //console.log("Effect took " + (t1 - t0) + " ms.");
    }

    finalize(type) {
        this.sourcectx.globalCompositeOperation = 'destination-in';
        this.sourcectx.drawImage(this.mask, 0, 0, this.mask.width, this.mask.height);
        //If inverse apply effect on bg layer
        if (this.reverse_mode_configuration.enabled &&
            (this.reverse_mode_configuration.process_no_result || this.effects_applied > 0)
        ) {
            let rect = null;
            switch (type) {
                case effects.PIXEL:
                    rect = new Rectangle(0, 0, this.natural_width, this.natural_height);
                    let box_size = pixel_configuration._pixel_size;
                    if (this.pixel_configuration._pixel_size_mode === 0) {
                        const smallest_dimension = Math.min(this.natural_height, this.natural_width);
                        box_size = Math.ceil(smallest_dimension / 100 * this.pixel_configuration._pixel_size);
                    }
                    if (this.pixel_configuration._geometry == 1) {
                        hexelate_area(this.marvin_out.ctx, rect, box_size, this.pixel_configuration._distance, false, this.marvin_out);
                    } else if (this.pixel_configuration._geometry == 2) {
                        hexelate_area(this.marvin_out.ctx, rect, box_size, this.pixel_configuration._distance, true, this.marvin_out);
                    } else {
                        pixelate_rectangle(this.marvin_out.ctx, rect, box_size, this.pixel_configuration._distance, this.marvin_out);
                    }
                    break;
                case effects.BLUR:
                    if (this.blurred == null) {
                        this.marvin_out.update();
                        this.blurred = generateBaseBlurImproved(this.marvin_out.canvas, this.blur_configuration._blur_strength);
                    }
                    this.marvin_out.ctx.drawImage(this.blurred, 0, 0, this.natural_width, this.natural_height);
                    break;
                case effects.BLACK:
                    this.marvin_out.ctx.fillStyle = bar_configuration._color;
                    this.marvin_out.ctx.fillRect(0, 0, this.natural_width, this.natural_height);
                    break;
                case effects.TRIANGLE:
                    let imageData = this.marvin_out.ctx.getImageData( 0, 0, this.source.width,  this.source.height );
                    const triangulize_configuration = this.triangle_configuration;
                    let triangulationParams = {
                        accuracy: triangulize_configuration.accuracy,
                        blur: triangulize_configuration.blur,
                        threshold: triangulize_configuration.threshold,
                        vertexCount: triangulize_configuration.vertexCount,
                        fill: triangulize_configuration.fill,
                        stroke: triangulize_configuration.stroke,
                        strokeWidth: triangulize_configuration.strokeWidth,
                        gradients: triangulize_configuration.gradients,
                        gradientStops: triangulize_configuration.gradientStops,
                        lineJoin: triangulize_configuration.lineJoin,
                        transparentColor: triangulize_configuration.transparentColor,
                    };
                    let triangulizedImageData = triangulate(triangulationParams)
                        .fromImageDataSync( imageData )
                        .toImageDataSync();
                    this.marvin_out.ctx.putImageData( triangulizedImageData, 0, 0 );
                    break;
                case effects.BOX:
                    //this.namedBoxRect(klass, score, rect);
                    break;

                case effects.GLITCH:
                    //this.fettyfiArea(klass, rect);
                    break;
                case effects.STICKER:
                    //??
                    break;
                case effects.NONE:
                    break;
            }
        }
        this.marvin_out.ctx.drawImage(this.source, 0, 0, this.mask.width, this.mask.height);
        this.marvin_out.imageData = this.marvin_out.ctx.getImageData(0, 0, this.natural_width, this.natural_height);

        //let t0 = performance.now();
        for (let layer_index = 0; layer_index < this.per_layer_queues.length; layer_index++) {
            if (layer_index in this.per_layer_queues) {
                let entries = this.per_layer_queues[layer_index];
                entries.forEach((entry) => {
                    let rect = entry[0];
                    let canvas = entry[1];
                    let mask = entry[2];
                    this.apply_section(rect, canvas, mask);
                });
            }
        }
        //let t1 = performance.now();
        //console.log("Effect layer_index took " + (t1 - t0) + " ms.");
        this.marvin_out.update();
    }

    apply_section(rect, canvas, mask) {
        //TODO: We should store the ctx or better the data object in the queue so we don't have to create it again and again
        const ctx = canvas.getContext('2d');
        // console.warn("Details for section apply")
        // console.log(rect);
        // console.log(canvas.width, canvas.height);
        // console.log(mask.calcWidth(), mask.calcHeight());
        // let failed = 0;
        // let success = 0;
        //
        // for (let x = 0; x < rect.w; x++) {
        //     for (let y = 0; y < rect.h; y++) {
        //         const fg_pixel = ctx.getImageData(x, y, 1, 1).data;
        //         const fr = Math.round(fg_pixel[0]);
        //         const fg = Math.round(fg_pixel[1]);
        //         const fb = Math.round(fg_pixel[2]);
        //
        //         if (fr === 0 && fg === 0 && fb === 0) {
        //             failed++;
        //         } else {
        //             success++;
        //         }
        //
        //         const fg_color = argb_to_int([0xff, fr, fg, 0xff]);
        //         this.marvin_out.setIntColor1(rect.x + x, rect.y + y, fg_color);
        //     }
        // }
        //
        // console.log("successes: " + success + " failures: " + failed);


        const ctx_data = ctx.getImageData(0, 0, canvas.width, canvas.height).data;

        for (let x in mask.container) {
            if (!mask.container.hasOwnProperty(x)) continue;
            // x is sourced from the eb_x which is an x from the origin of the eb_rectangle
            x = parseInt(x);

            for (let y in mask.container[x]) {
                if (!mask.container[x].hasOwnProperty(y)) continue;
                y = parseInt(y);

                const coordinate_mask = mask.container[x][y];
                /*
                const fg_pixel = ctx.getImageData(x, y, 1, 1).data;
                const fr = Math.round(fg_pixel[0]);
                const fg = Math.round(fg_pixel[1]);
                const fb = Math.round(fg_pixel[2]);
                */
                const index = (x + y * canvas.width) * 4;
                const fr = ctx_data[index];
                const fg = ctx_data[index + 1];
                const fb = ctx_data[index + 2];
                //const alpha = ctx_data[index+3];

                //Many calculations happen twice here and on top of that we convert colors to different representations
                //just for them to be converted back again as the method just wants the color in this format as the parameter
                //Info: The loop entire itself takes 500ms (without code in it), while these 3 code lines "only" consume 300ms though
                // const fg_color = argb_to_int([0xff, fr, fg, fb]);
                // this.marvin_out.setIntColor1(rect.x + x, rect.y + y, fg_color);

                const fg_color = argb_to_int([0xff, fr, fg, fb]) & coordinate_mask;
                const bg_color = this.marvin_out.getIntColor(rect.x + x, rect.y + y) & ~coordinate_mask;
                this.marvin_out.setIntColor1(rect.x + x, rect.y + y, fg_color | bg_color);


                // this.marvin_out.setIntColor1(rect.x + x, rect.y + y, coordinate_mask);
            }
        }


    }

    neo_shift(source_context, source_data, target_context, target_canvas, rect, effect_bounds, config) {
        // TODO implement internal H/V distinction.
        let shift_range = config.shift_intensity_range;
        const hShifts = this.calculateShifts(rect.h, config.h_bar_size_range, shift_range);
        const vShifts = this.calculateShifts(rect.w, config.v_bar_size_range, shift_range);

        // TODO per block
        const mask = new Mask();

        // const rect_data_obj = source_context.getImageData(0, 0, rect.w, rect.h);
        // const rect_data = rect_data_obj.data;
        const image_data_obj = source_context.getImageData(0, 0, effect_bounds.w, effect_bounds.h);
        const image_data = image_data_obj.data;

        let i = 0;
        for (let rect_y = 0; rect_y < rect.h; rect_y++) {
            const hShift = hShifts[rect_y];
            if (hShift === undefined) {
                // console.warn("Aborting early as the remaining pixels (" + (rect.h - y) + ") is smaller than the minimum allowed bar height.");
                break;
            }
            const eb_y = rect_y + (rect.y - effect_bounds.y);

            for (let rect_x = 0; rect_x < rect.w; rect_x++) {

                const vShift = vShifts[rect_x];
                if (vShift === undefined) {
                    // console.warn("Aborting early as the remaining pixels (" + (rect.w - x) + ") is smaller than the minimum allowed bar width.");
                    break;
                }
                const eb_x = rect_x + (rect.x - effect_bounds.x);

                // const a = source.getAlphaComponent(x, y);
                //const source_pixel = source_context.getImageData(eb_x, eb_y, 1, 1).data;
                const index = (eb_x + eb_y * source_data.width) * 4;
                const red = source_data.data[index];
                const green = source_data.data[index + 1];
                const blue = source_data.data[index + 2];
                const alpha = source_data.data[index + 3];
                //const source_pixel = rect_data[];
                const sr = Math.round(red * config.r_channel_intensity);
                const sg = Math.round(green * config.g_channel_intensity);
                const sb = Math.round(blue * config.b_channel_intensity);
                const tx = eb_x + hShift;
                const ty = eb_y + vShift;

                // if (!effect_bounds.coordinateInside(tx, ty)) continue;
                /*
                let target_pixel;
                try {
                    target_pixel = source_context.getImageData(tx, ty, 1, 1).data;
                } catch (e) {
                    console.error(rect_x, rect_y, eb_x, eb_y, hShift, vShift, tx, ty, rect);
                    return null;
                }

                const tr = Math.round(target_pixel[0] * (1 - config.r_channel_intensity));
                const tg = Math.round(target_pixel[1] * (1 - config.g_channel_intensity));
                const tb = Math.round(target_pixel[2] * (1 - config.b_channel_intensity));
                const ta = alpha;
                */
                let t_index = (tx + ty * source_data.width) * 4;
                const tr = Math.round(source_data.data[t_index] * (1 - config.r_channel_intensity));
                const tg = Math.round(source_data.data[t_index + 1] * (1 - config.g_channel_intensity));
                const tb = Math.round(source_data.data[t_index + 2] * (1 - config.b_channel_intensity));
                // const tr = 0;
                // const tg = 0;
                // const tb = 0;
                const ta = alpha;

                if (config.r_channel_intensity > 0) mask.setRed(tx, ty);
                if (config.g_channel_intensity > 0) mask.setGreen(tx, ty);
                if (config.b_channel_intensity > 0) mask.setBlue(tx, ty);
                /*
                const image_data = target_context.createImageData(1, 1);
                const id_bytes = image_data.data;
                id_bytes[0] = sr + tr;
                id_bytes[1] = sg + tg;
                id_bytes[2] = sb + tb;
                id_bytes[3] = ta;
                target_context.putImageData(image_data, tx, ty);
                */

                t_index = (tx + ty * image_data_obj.width) * 4;
                image_data_obj.data[t_index] = sr + tr;
                image_data_obj.data[t_index + 1] = sg + tg;
                image_data_obj.data[t_index + 2] = sb + tb;
                image_data_obj.data[t_index + 3] = ta;

            }
        }
        target_context.putImageData(image_data_obj, 0, 0);
        return mask;
    }

    calculateShifts(dimension, bar_range, shift_range) {
        if (bar_range[1] === 0 && bar_range[0] === 0) {
            return new Array(dimension).fill(0);
        }

        let bar_size = 0;
        let shift_intensity = 0;
        const indexed_shifts = [];
        for (let i = 0; i < dimension; i++) {
            if (bar_size > 0) {
                bar_size--;
            } else {
                if (dimension - i < bar_range[1]) {
                    // console.warn("Breaking loop as " + dimension + " - " + i + " < " + bar_range[1] + ".");
                    break;
                }
                bar_size = generate_bar_size(bar_range[0], bar_range[1]);
                shift_intensity = generate_shift_intensity(shift_range[0], shift_range[1], shift_intensity);
            }
            indexed_shifts[i] = shift_intensity;
        }
        return indexed_shifts;
    }
}

function int_to_argb(int_color) {
    const a_channel = (int_color >> 24) & 0xFF;
    const r_channel = (int_color >> 16) & 0xFF;
    const g_channel = (int_color >> 8) & 0xFF;
    const b_channel = int_color & 0xFF;
    return [a_channel, r_channel, g_channel, b_channel]
}

function argb_to_int(channels) {
    const a_channel = (channels[0] & 0xFF) << 24;
    const r_channel = (channels[1] & 0xFF) << 16;
    const g_channel = (channels[2] & 0xFF) << 8;
    const b_channel = (channels[3] & 0xFF);
    return a_channel | r_channel | g_channel | b_channel;
}

function generate_bar_size(min, max) {
    const random_value = Math.random() * (max - min);
    return Math.floor(min + random_value);
}

function generate_shift_intensity(min, max, previous) {
    const r_base = min + Math.round(Math.random() * (max - min));
    const r_sign = Math.round(Math.random()) === 0 ? -1 : 1;
    const random_delta = r_base * r_sign;

    let candidate_shift = previous + random_delta;
    if (candidate_shift < -max || max < candidate_shift) {
        candidate_shift = previous - random_delta;
    }
    return candidate_shift;
}

function make_region_buffer(local_image, x_origin, y_origin, width, height) {
    const buffer = new Array(width * height);

    for (let y_local = 0; y_local < height; y_local++) {
        let y_global = y_origin + y_local;
        for (let x_local = 0; x_local < width; x_local++) {
            let x_global = x_origin + x_local;
            buffer[y_local * width + x_local] = local_image.getIntColor(x_global, y_global)
        }
    }
    return buffer;
}

function calculate_combined_pixel(original, x_global, y_global, x_shift, y_shift, config) {
    const original_pixel = original.getIntColor(x_global, y_global);
    const original_argb = int_to_argb(original_pixel);

    const shifted_pixel = original.getIntColor(x_global + x_shift, y_global + y_shift)
    const shifted_argb = int_to_argb(shifted_pixel);

    // noinspection DuplicatedCode
    return actually_combine_pixel_values(original_argb, shifted_argb, config);
}

function actually_combine_pixel_values(original_argb, shifted_argb, config) {
    const combined_argb = [
        original_argb[0],
        shifted_argb[1] * config.r_channel_intensity + original_argb[1] * (1 - config.r_channel_intensity),
        shifted_argb[2] * config.g_channel_intensity + original_argb[2] * (1 - config.g_channel_intensity),
        shifted_argb[3] * config.b_channel_intensity + original_argb[3] * (1 - config.b_channel_intensity),
    ];

    // noinspection DuplicatedCode
    const negative_combined_argb = [
        original_argb[0],
        (0xFF - shifted_argb[1]) * config.r_channel_intensity + (0xFF - original_argb[1]) * (1 - config.r_channel_intensity),
        (0xFF - shifted_argb[2]) * config.g_channel_intensity + (0xFF - original_argb[2]) * (1 - config.g_channel_intensity),
        (0xFF - shifted_argb[3]) * config.b_channel_intensity + (0xFF - original_argb[3]) * (1 - config.b_channel_intensity),
    ];

    // noinspection DuplicatedCode
    const ultimate_combined_argb = [
        combined_argb[0],
        negative_combined_argb[1] * config.r_negate_intensity + combined_argb[1] * (1 - config.r_negate_intensity),
        negative_combined_argb[2] * config.g_negate_intensity + combined_argb[2] * (1 - config.g_negate_intensity),
        negative_combined_argb[3] * config.b_negate_intensity + combined_argb[3] * (1 - config.b_negate_intensity),
    ];

    return argb_to_int(ultimate_combined_argb);
}

function merge_pixels(bg_pixel, fg_pixel) {

    const fg_argb = int_to_argb(fg_pixel);
    const bg_argb = int_to_argb(bg_pixel);

    const fg_factor = (fg_argb[0] / 255.);
    const bg_factor = 1 - fg_factor;

    const combined_argb = [
        bg_argb[0],
        fg_argb[1] & 0b1 ? bg_argb[1] * bg_factor + fg_argb[1] * fg_factor : bg_argb[1],
        fg_argb[2] & 0b1 ? bg_argb[2] * bg_factor + fg_argb[2] * fg_factor : bg_argb[2],
        fg_argb[3] & 0b1 ? bg_argb[3] * bg_factor + fg_argb[3] * fg_factor : bg_argb[3],
    ];

    return argb_to_int(combined_argb);
}

function generateBaseBlurImproved(canvas, blur_strength) {
    //marvinImage.update();
    let blurredcanvas = document.createElement('canvas');
    let pxext = blur_strength * 2;
    let pxext2 = pxext * 2;
    blurredcanvas.width = canvas.width + pxext2;
    blurredcanvas.height = canvas.height + pxext2;
    const ctx = blurredcanvas.getContext('2d');
    ctx.save();
    //Upper left corner
    ctx.translate(pxext, pxext);
    ctx.scale(-1, -1);
    ctx.drawImage(canvas,
        0, 0, pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Top
    ctx.translate(0, pxext);
    ctx.scale(1, -1);
    ctx.drawImage(canvas,
        0, 0, canvas.width, pxext,
        pxext, 0, canvas.width, pxext);
    ctx.restore();
    ctx.save();
    //Upper right corner
    ctx.translate(blurredcanvas.width, pxext);
    ctx.scale(-1, -1);
    ctx.drawImage(canvas,
        canvas.width - pxext, 0, pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Left
    ctx.translate(pxext, 0);
    ctx.scale(-1, 1);
    ctx.drawImage(canvas,
        0, 0, pxext, canvas.height,
        0, pxext, pxext, canvas.height
        //0, 0, marvinImage.width, marvinImage.height
    );
    ctx.restore();
    ctx.save();
    //Right
    ctx.translate(blurredcanvas.width, 0);
    ctx.scale(-1, 1);
    ctx.drawImage(canvas,
        canvas.width - pxext, 0, pxext, canvas.height,
        0, pxext, pxext, canvas.height);
    ctx.restore();
    ctx.save();
    //Bottom left corner
    ctx.translate(pxext, blurredcanvas.height);
    ctx.scale(-1, -1);
    ctx.drawImage(canvas,
        0, canvas.height - pxext, pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Bottom
    ctx.translate(0, blurredcanvas.height);
    ctx.scale(1, -1);
    ctx.drawImage(canvas,
        0, canvas.height - pxext, canvas.width, pxext,
        pxext, 0, canvas.width, pxext);
    ctx.restore();
    ctx.save();
    //Bottom right corner
    ctx.translate(blurredcanvas.width, blurredcanvas.height);
    ctx.scale(-1, -1);
    ctx.drawImage(canvas,
        canvas.width - pxext, canvas.height - pxext, pxext, pxext,
        0, 0, pxext, pxext);
    ctx.restore();
    ctx.save();
    //Center
    ctx.drawImage(canvas, pxext, pxext, canvas.width, canvas.height);
    ctx.filter = 'blur(' + blur_strength + 'px)';
    ctx.drawImage(blurredcanvas, 0, 0, blurredcanvas.width, blurredcanvas.height);
    let blurred = document.createElement('canvas');
    blurred.width = canvas.width;
    blurred.height = canvas.height;
    const ctx_blurred = blurred.getContext('2d');

    ctx_blurred.drawImage(blurredcanvas,
        pxext, pxext, canvas.width, canvas.height,
        0, 0, canvas.width, canvas.height);

    return blurred;
}

function pixelate_rectangle(ctx, rect, pixel_size = 20, distance = 0, marvinImage) {
    //marvinImage.update();
    let pixelArr = ctx.getImageData(0, 0, marvinImage.getWidth(), marvinImage.getHeight()).data;
    for (let i = rect.x; i < rect.x + rect.w; i += pixel_size + distance) {
        for (let j = rect.y; j < rect.y + rect.h; j += pixel_size + distance) {
            ctx.fillStyle = calculatePixelBoxColor(marvinImage, i, j, i + pixel_size, j + pixel_size, pixel_size, pixelArr);
            ctx.fillRect(i, j, pixel_size, pixel_size);
        }
    }
    //marvinImage.imageData =  marvinImage.ctx.getImageData(0, 0, marvinImage.getWidth(), marvinImage.getHeight());
    //return marvinImage;
}

function toColor(num, alpha = false) {
    num >>>= 0;
    let b = num & 0xFF,
        g = (num & 0xFF00) >>> 8,
        r = (num & 0xFF0000) >>> 16,
        a = ((num & 0xFF000000) >>> 24) / 255;
    if (!alpha) {
        a = 1;
    }
    return "rgba(" + [r, g, b, a].join(",") + ")";
}

function hexelate_area(ctx, rect, pixel_size = 20, distance = 0, glitch = false, marvinImage) {
    let a = 2 * Math.PI / 6;
    let r = pixel_size;
    if (glitch) {
        a = Math.ceil(a);
    }

    function drawHexagon(ctx, x, y, r, a, pixelArr) {

        ctx.beginPath();
        for (let i = 0; i < 6; i++) {
            ctx.lineTo(x + r * Math.cos(a * i), y + r * Math.sin(a * i));
        }
        ctx.fillStyle = calculatePixelBoxColor(marvinImage, Math.ceil(x), Math.ceil(y), 0, 0, r, pixelArr);

        ctx.fill();
        ctx.closePath();
    }

    //marvinImage.update();
    let pixelArr = ctx.getImageData(0, 0, marvinImage.getWidth(), marvinImage.getHeight()).data;

    //for (let y = rect.y + r; y + r * Math.sin(a) < rect.y + rect.h; y += 2 * (r * Math.sin(a))) {
    for (let y = rect.y; y < rect.y + rect.h + r; y += 2 * (r * Math.sin(a))) {
        y += distance;
        //for (let x = rect.x + r, j = 0; x + r * (1 + Math.cos(a)) < rect.x + rect.w; x += r * (1 + Math.cos(a))) {
        for (let x = rect.x, j = 0; x < rect.x + rect.w + r; x += r * (1 + Math.cos(a))) {
            x += distance;
            let yc = y + 0.5 * ((-1) ** j++ * r * Math.sin(a));
            drawHexagon(ctx, x, yc, r, a, pixelArr);
        }
    }
    //marvinImage.imageData = marvinImage.ctx.getImageData(0, 0, marvinImage.width, marvinImage.height);
    //marvinImage.update();
    //return marvinImage;
}

function calc_smoothing(rect, config) {
    let smoothing_strength = Math.ceil(Math.min(rect.w, rect.h) / 100 * config._feathering);
    return smoothing_strength;
}

/**
 *
 * @param marvinImage
 * @param x1 = pixel box starting x
 * @param y1 = pixel box starting y
 * @param x2 = pixel box ending x
 * @param y2 = pixel box ending y
 * @param pixel_size = box size
 * @param pixelArr = pixel array
 * @returns {*}
 */
function calculatePixelBoxColor(marvinImage, x1, y1, x2, y2, pixel_size, pixelArr) {
    let p = (x1 + (y1 * marvinImage.getWidth())) * 4;
    return "rgba(" + pixelArr[p] + "," + pixelArr[p + 1] + "," + pixelArr[p + 2] + "," + pixelArr[p + 3] + ")";
    /* These had a bad performance
    switch (1) {
        case 0:
            return marvinImage.getIntColor(x1, y1);
        case 1:
            let rectarea = new MarvinImage();
            let pixelmarvin = new MarvinImage();
            Marvin.crop(marvinImage, rectarea, x1, y1, pixel_size, pixel_size);
            Marvin.scale(rectarea, pixelmarvin, 1, 1);
            return pixelmarvin.getIntColor(0, 0);
        default:
    }
    return marvinImage.getIntColor(x1, y1);
    */
}